@@ -1,15 +1,13 @@
-from __future__ import generator_stop
-
 import collections
 import itertools
 import os
 import re
 import shutil
-import tempfile
 
+from dulwich.client import HTTPUnauthorized
 from dulwich.errors import HangupException, GitProtocolError, ApplyDeltaError
 from dulwich.objects import Blob, Commit, Tag, Tree, parse_timezone
-from dulwich.pack import create_delta, apply_delta
+from dulwich.pack import apply_delta
 from dulwich.refs import (
     ANNOTATED_TAG_SUFFIX,
     LOCAL_BRANCH_PREFIX,
@@ -21,8 +19,8 @@ from dulwich import config as dul_config
 from dulwich import diff_tree
 
 from mercurial.i18n import _
-from mercurial.node import hex, bin, nullid, short
-from mercurial.utils import dateutil
+from mercurial.node import hex, bin, nullid, nullhex, short
+from mercurial.utils import dateutil, urlutil
 from mercurial import (
     bookmarks,
     context,
@@ -88,7 +86,7 @@ class GitProgress(object):
         # 'Counting objects: 33640, done.\n'
         # 'Compressing objects:   0% (1/9955)   \r
 
-        lines = (self.msgbuf + compat.sysbytes(message)).splitlines(
+        lines = (self.msgbuf + pycompat.sysbytes(message)).splitlines(
             keepends=True
         )
         self.msgbuf = b''
@@ -130,9 +128,11 @@ class GitProgress(object):
 
             if determinate:
                 pos, total = map(int, determinate.group(1, 2))
-            else:
+            elif indeterminate:
                 pos = int(indeterminate.group(1))
                 total = None
+            else:
+                continue
 
             self._progress.update(pos, total=total)
 
@@ -143,8 +143,25 @@ class GitProgress(object):
         self.progress(b'')
 
 
+class heads_tags(object):
+    __slots__ = "heads", "tags"
+
+    def __init__(self, heads=(), tags=()):
+        self.heads = set(heads)
+        self.tags = set(tags)
+
+    def __iter__(self):
+        return itertools.chain(self.heads, self.tags)
+
+    def __bool__(self):
+        return bool(self.heads) or bool(self.tags)
+
+    def __repr__(self):
+        return f"heads_tags(heads={self.heads}, tags={self.tags})"
+
+
 def get_repo_and_gitdir(repo):
-    if repo.shared():
+    if repo.local() and repo.shared():
         repo = hg.sharedreposource(repo)
 
     if repo.ui.configbool(b'git', b'intree'):
@@ -156,7 +173,7 @@ def get_repo_and_gitdir(repo):
 
 
 def has_gitrepo(repo):
-    if not hgutil.safehasattr(repo, b'vfs'):
+    if not hasattr(repo, 'vfs'):
         return False
 
     repo, gitdir = get_repo_and_gitdir(repo)
@@ -269,12 +286,13 @@ class GitHandler(object):
     def map_hg_get(self, gitsha, deref=False):
         if deref:
             try:
-                gitobj = self.git.get_object(gitsha)
+                unpeeled, peeled = compat.peel_sha(
+                    self.git.object_store, gitsha
+                )
+                gitsha = peeled.id
             except KeyError:
-                gitobj = None
-
-            if isinstance(gitobj, Tag):
-                gitsha = gitobj.object[1]
+                self.ui.note(b'note: failed to dereference %s\n' % gitsha)
+                return None
 
         return self._map_git.get(gitsha)
 
@@ -314,7 +332,8 @@ class GitHandler(object):
         if os.path.exists(self.vfs.join(self.tags_file)):
             for line in self.vfs(self.tags_file):
                 sha, name = line.strip().split(b' ', 1)
-                self.tags[name] = sha
+                if sha in self.repo.unfiltered():
+                    self.tags[name] = sha
 
     def save_tags(self):
         with self.repo.lock():
@@ -325,33 +344,33 @@ class GitHandler(object):
 
     def load_remote_refs(self):
         self._remote_refs = {}
-        refdir = os.path.join(self.gitdir, b'refs', b'remotes')
+
+        # don't do anything if there's no git repository, as accessing
+        # `self.git` will create it
+        if not os.path.isdir(self.gitdir):
+            return
 
         # if no paths are set, we should still check 'default'
         pathnames = list(self.ui.paths) or [b'default']
 
-        # we avoid using dulwich's refs method because it is incredibly slow;
-        # on a repo with a few hundred branches and a few thousand tags,
-        # dulwich took about 200ms
         for pathname in pathnames:
-            remotedir = os.path.join(refdir, pathname)
-            for root, dirs, files in os.walk(remotedir):
-                for f in files:
-                    try:
-                        ref = root.replace(refdir + pycompat.ossep, b'') + b'/'
-                        node = open(os.path.join(root, f), 'rb').read().strip()
-                        self._remote_refs[ref + f] = bin(self._map_git[node])
-                    except (KeyError, IOError):
-                        pass
+            base = b'refs/remotes/%s/' % pathname
+            for ref in self.git.refs.subkeys(base):
+                ref = base + ref
+                sha = self.git.refs[ref]
+                if sha in self._map_git:
+                    node = bin(self._map_git[sha])
+                    if node in self.repo.unfiltered():
+                        self._remote_refs[ref[13:]] = node
 
     # END FILE LOAD AND SAVE METHODS
 
     # COMMANDS METHODS
 
     def import_commits(self, remote_name):
+        remote_names = [remote_name] if remote_name is not None else []
         refs = self.git.refs.as_dict()
-        remote_names = [remote_name] if remote_name else []
-        self.import_git_objects(remote_names, refs)
+        self.import_git_objects(b'gimport', remote_names, refs)
 
     def fetch(self, remote, heads):
         result = self.fetch_pack(remote.path, heads)
@@ -361,6 +380,7 @@ class GitHandler(object):
 
         if result.refs:
             imported = self.import_git_objects(
+                b'pull',
                 remote_names,
                 result.refs,
                 heads=heads,
@@ -474,18 +494,20 @@ class GitHandler(object):
                 _(b"git remote error: ") + pycompat.sysbytes(str(e))
             )
 
-    def push(self, remote, revs, force):
-        old_refs, new_refs = self.upload_pack(remote, revs, force)
-        remote_names = self.remote_names(remote, True)
-        remote_desc = remote_names[0] if remote_names else remote
+    def push(self, remote, revs, bookmarks, force):
+        self.repo.hook(
+            b"preoutgoing",
+            git=True,
+            source=b'push',
+            url=remote,
+        )
 
-        if not isinstance(new_refs, dict):
-            # dulwich 0.20.6 changed the API and deprectated treating
-            # the result as a dictionary
-            ref_status = new_refs.ref_status
-            new_refs = new_refs.refs
-        else:
-            ref_status = {}
+        old_refs, new_refs = self.upload_pack(remote, revs, bookmarks, force)
+        remote_names = self.remote_names(remote, True)
+        remote_desc = remote_names[0] if remote_names else b''
+
+        ref_status = new_refs.ref_status
+        new_refs = new_refs.refs
 
         for ref, new_sha in sorted(new_refs.items()):
             old_sha = old_refs.get(ref)
@@ -494,6 +516,8 @@ class GitHandler(object):
                     b'warning: failed to update %s; %s\n'
                     % (ref, pycompat.sysbytes(ref_status[ref])),
                 )
+            elif new_sha == nullhex:
+                self.ui.status(b"deleting reference %s\n" % ref)
             elif old_sha is None:
                 if self.ui.verbose:
                     self.ui.note(
@@ -525,11 +549,10 @@ class GitHandler(object):
                 new_refs_with_head.update(
                     self.fetch_pack(remote, [b'HEAD']).refs,
                 )
-            except (error.RepoLookupError):
+            except error.RepoLookupError:
                 self.ui.debug(b'remote repository has no HEAD\n')
 
-            for remote_name in remote_names:
-                self.update_remote_branches(remote_name, new_refs_with_head)
+            self.update_remote_branches(remote_names, new_refs_with_head)
 
         if old_refs == new_refs:
             if revs or not old_refs:
@@ -598,7 +621,12 @@ class GitHandler(object):
         else:
             reqrefs = result.refs
 
-        commits = [c.node for c in self.get_git_incoming(reqrefs)]
+        commits = [
+            c.node
+            for c in self.get_git_incoming(
+                reqrefs, self.remote_names(remote.path, push=False)
+            )
+        ]
 
         b = overlayrepo(self, commits, result.refs)
 
@@ -633,6 +661,8 @@ class GitHandler(object):
 
         self.ui.note(_(b"exporting %d changesets\n") % total)
 
+        self.repo.hook(b'gitexport', nodes=[c.hex() for c in export], git=True)
+
         # By only exporting deltas, the assertion is that all previous objects
         # for all other changesets are already present in the Git repository.
         # This assertion is necessary to prevent redundant work. Here, nodes,
@@ -725,14 +755,14 @@ class GitHandler(object):
                 commit.parents.append(git_sha)
 
         commit.message, extra = self.get_git_message_and_extra(ctx)
-        commit.extra.extend(extra)
+        commit._extra.extend(extra)
 
         if b'encoding' in extra:
             commit.encoding = extra[b'encoding']
         if b'gpgsig' in extra:
             commit.gpgsig = extra[b'gpgsig']
 
-        for obj, nodeid in exporter.update_changeset(ctx):
+        for obj in exporter.update_changeset(ctx):
             if obj.id not in self.git.object_store:
                 self.git.object_store.add_object(obj)
 
@@ -830,7 +860,7 @@ class GitHandler(object):
         if b'author' in ctx.extra():
             try:
                 author = b"".join(apply_delta(author, ctx.extra()[b'author']))
-            except ApplyDeltaError:
+            except (ApplyDeltaError, AssertionError):
                 self.ui.traceback()
                 self.ui.warn(
                     b"warning: disregarding possibly invalid metadata in %s\n"
@@ -866,7 +896,7 @@ class GitHandler(object):
         if b'message' in extra:
             try:
                 message = b"".join(apply_delta(message, extra[b'message']))
-            except ApplyDeltaError:
+            except (ApplyDeltaError, AssertionError):
                 self.ui.traceback()
                 self.ui.warn(
                     b"warning: disregarding possibly invalid metadata in %s\n"
@@ -964,12 +994,13 @@ class GitHandler(object):
 
         return message, git_extra
 
-    def get_git_incoming(self, refs):
+    def get_git_incoming(self, refs, remote_names):
         return git2hg.find_incoming(
             self.ui,
             self.git.object_store,
             self._map_git,
             refs,
+            remote_names,
         )
 
     def get_transaction(self, desc=b"hg-git"):
@@ -1004,11 +1035,18 @@ class GitHandler(object):
         else:
             return None, None
 
-    def import_git_objects(self, remote_names, refs, heads=None):
-        filteredrefs = self.filter_min_date(refs)
-        if heads is not None:
-            filteredrefs = git2hg.filter_refs(filteredrefs, heads)
-        commits = self.get_git_incoming(filteredrefs)
+    def import_git_objects(self, command, remote_names, refs, heads=None):
+        self.repo.hook(
+            b'gitimport',
+            source=command,
+            git=True,
+            names=remote_names,
+            refs=refs,
+            heads=heads,
+        )
+
+        filteredrefs = git2hg.filter_refs(self.filter_min_date(refs), heads)
+        commits = self.get_git_incoming(filteredrefs, remote_names)
         # import each of the commits, oldest first
         total = len(commits)
         if total:
@@ -1036,23 +1074,38 @@ class GitHandler(object):
             # get at least one chunk
             for offset in range(0, max(total, 1), chunksize):
                 with self.get_transaction(b"gimport"):
+                    oldtiprev = self.repo.changelog.tiprev()
+
                     for commit in commits[offset : offset + chunksize]:
                         progress.increment(item=commit.short)
                         self.import_git_commit(
+                            command,
                             self.git[commit.sha],
                             commit.phase,
                         )
 
+                    lastrev = self.repo.changelog.tiprev()
+
                     self.import_tags(refs)
                     self.update_hg_bookmarks(remote_names, refs)
+                    self.update_remote_branches(remote_names, refs)
 
-                    for remote_name in remote_names:
-                        self.update_remote_branches(remote_name, refs)
+                    if oldtiprev != lastrev:
+                        first = self.repo.changelog.node(oldtiprev + 1)
+                        last = self.repo.changelog.node(lastrev)
+
+                        self.repo.hook(
+                            b"changegroup",
+                            source=b'push',
+                            git=True,
+                            node=hex(first),
+                            node_last=hex(last),
+                        )
 
         # TODO if the tags cache is used, remove any dangling tag references
         return total
 
-    def import_git_commit(self, commit, phase):
+    def import_git_commit(self, command, commit, phase):
         self.ui.debug(_(b"importing: %s\n") % commit.id)
         unfiltered = self.repo.unfiltered()
 
@@ -1062,7 +1115,7 @@ class GitHandler(object):
             hg_renames,
             hg_branch,
             extra,
-        ) = git2hg.extract_hg_metadata(commit.message, commit.extra)
+        ) = git2hg.extract_hg_metadata(commit.message, commit._extra)
         if hg_renames is None:
             detect_renames = True
             # We have to store this unconditionally, even if there are no
@@ -1143,9 +1196,12 @@ class GitHandler(object):
             hgsub = util.parse_hgsub(
                 git2hg.git_file_readlines(self.git, git_commit_tree, b'.hgsub')
             )
-            for (sm_path, sm_url, sm_name) in gitmodules:
+            for sm_path, sm_url, sm_name in gitmodules:
                 hgsub[sm_path] = b'[git]' + sm_url
+            for path in hgsubstate.keys() - hgsub.keys():
+                del hgsubstate[path]
             files[b'.hgsub'] = (False, 0o100644, None)
+            files.pop(b'.gitmodules', None)
         elif (
             commit.parents
             and b'.gitmodules' in self.git[self.git[commit.parents[0]].tree]
@@ -1166,7 +1222,7 @@ class GitHandler(object):
 
         text = b'\n'.join(l.rstrip() for l in text.splitlines()).strip(b'\n')
         if text + b'\n' != origtext:
-            extra[b'message'] = create_delta(text + b'\n', origtext)
+            extra[b'message'] = util.create_delta(text + b'\n', origtext)
 
         author = commit.author
 
@@ -1187,7 +1243,7 @@ class GitHandler(object):
         except UnicodeDecodeError:
             origauthor = author
             author = util.decode_guess(author, commit.encoding)
-            extra[b'author'] = create_delta(author, origauthor)
+            extra[b'author'] = util.create_delta(author, origauthor)
 
         oldenc = util.swap_out_encoding()
 
@@ -1315,6 +1371,7 @@ class GitHandler(object):
             date,
             extra,
         )
+
         # Starting Mercurial commit d2743be1bb06, memctx imports from
         # committablectx. This means that it has a 'substate' property that
         # contains the subrepo state. Ordinarily, Mercurial expects the subrepo
@@ -1340,9 +1397,23 @@ class GitHandler(object):
         cs = hex(node)
         self.map_set(commit.id, cs)
 
+        self.repo.hook(
+            b'incoming',
+            git=True,
+            source=command,
+            node=cs,
+            git_node=commit.id,
+        )
+
     # PACK UPLOADING AND FETCHING
 
-    def upload_pack(self, remote, revs, force):
+    def upload_pack(self, remote, revs, bookmarks, force):
+        if bookmarks and self.branch_bookmark_suffix:
+            raise error.Abort(
+                b"the -B/--bookmarks option is not supported when "
+                b"branch_bookmark_suffix is set",
+            )
+
         all_exportable = self.export_commits()
         old_refs = {}
         change_totals = {}
@@ -1355,29 +1426,68 @@ class GitHandler(object):
             else:
                 exportable = {}
                 for rev in (hex(r) for r in revs):
-                    if rev not in all_exportable:
+                    if rev == nullhex:
+                        # a deletion
+                        exportable[rev] = heads_tags(
+                            heads={
+                                LOCAL_BRANCH_PREFIX + bm
+                                for bm in bookmarks
+                                if bm not in self.repo._bookmarks
+                            }
+                        )
+                    elif rev not in all_exportable:
                         raise error.Abort(
                             b"revision %s cannot be pushed since"
                             b" it doesn't have a bookmark" % self.repo[rev]
                         )
-                    exportable[rev] = all_exportable[rev]
-            return self.get_changed_refs(refs, exportable, force)
+                    elif bookmarks:
+                        # we should only push the listed bookmarks,
+                        # and not any other bookmarks that might point
+                        # to the same changeset
+                        exportable[rev] = heads_tags(
+                            heads=all_exportable[rev].heads
+                            & {LOCAL_BRANCH_PREFIX + bm for bm in bookmarks},
+                        )
+                    else:
+                        exportable[rev] = all_exportable[rev]
+
+            changes = self.get_changed_refs(refs, exportable, force)
+
+            self.repo.hook(
+                b"prechangegroup",
+                source=b'push',
+                git=True,
+                url=remote,
+                changes=changes,
+            )
+
+            return changes
 
         def genpack(have, want, progress=None, ofs_delta=True):
             commits = []
 
             with util.abort_push_on_keyerror():
-                missing = self.git.object_store.find_missing_objects(
+                for sha, name in compat.MissingObjectFinder(
+                    self.git.object_store,
                     have,
                     want,
-                )
+                    progress=progress,
+                ):
+                    o = self.git.object_store[sha]
+                    t = type(o)
+                    change_totals[t] = change_totals.get(t, 0) + 1
+                    if isinstance(o, Commit):
+                        commits.append(sha)
+
+                        self.repo.hook(
+                            b"outgoing",
+                            source=b'push',
+                            git=True,
+                            url=remote,
+                            node=self.map_hg_get(sha),
+                            git_node=sha,
+                        )
 
-            for sha, name in missing:
-                o = self.git.object_store[sha]
-                t = type(o)
-                change_totals[t] = change_totals.get(t, 0) + 1
-                if isinstance(o, Commit):
-                    commits.append(sha)
             commit_count = len(commits)
             self.ui.note(_(b"%d commits found\n") % commit_count)
             if commit_count > 0:
@@ -1420,7 +1530,7 @@ class GitHandler(object):
     def get_changed_refs(self, refs, exportable, force):
         new_refs = refs.copy()
 
-        if not exportable:
+        if not any(exportable.values()):
             raise error.Abort(
                 b'no bookmarks or tags to push to git',
                 hint=b'see "hg help bookmarks" for details on creating them',
@@ -1430,11 +1540,6 @@ class GitHandler(object):
         unfiltered = self.repo.unfiltered()
         for rev, rev_refs in exportable.items():
             ctx = self.repo[rev]
-            if not rev_refs:
-                raise error.Abort(
-                    b"revision %s cannot be pushed since"
-                    b" it doesn't have a bookmark" % ctx
-                )
 
             # Check if the tags the server is advertising are annotated tags,
             # by attempting to retrieve it from the our git repo, and building
@@ -1460,12 +1565,33 @@ class GitHandler(object):
                 uptodate_annotated_tags.append(ref)
 
             for ref in rev_refs:
-                if ref not in refs:
-                    gitobj = self.git.get_object(self.git.refs[ref])
-                    if isinstance(gitobj, Tag):
-                        new_refs[ref] = gitobj.id
+                if ctx.node() == nullid:
+                    if ref not in new_refs:
+                        # this is reasonably consistent with
+                        # mercurial; git aborts with an error in this
+                        # case
+                        self.ui.warn(
+                            b"warning: unable to delete '%s' as it does not "
+                            b"exist on the remote repository\n" % ref,
+                        )
                     else:
-                        new_refs[ref] = self.map_git_get(ctx.hex())
+                        new_refs[ref] = nullhex
+                elif (
+                    not util.ref_exists(ref, self.git.refs)
+                    and ref not in new_refs
+                ):
+                    self.ui.warn(b"warning: cannot update '%s'\n" % ref)
+                elif ref not in refs:
+                    if ref not in self.git.refs:
+                        self.ui.note(
+                            b'note: cannot update %s\n' % (ref),
+                        )
+                    else:
+                        gitobj = self.git.get_object(self.git.refs[ref])
+                        if isinstance(gitobj, Tag):
+                            new_refs[ref] = gitobj.id
+                        else:
+                            new_refs[ref] = self.map_git_get(ctx.hex())
                 elif new_refs[ref] in self._map_git:
                     rctx = unfiltered[self.map_hg_get(new_refs[ref])]
                     if rctx.ancestor(ctx) == rctx or force:
@@ -1503,44 +1629,16 @@ class GitHandler(object):
 
         progress = GitProgress(self.ui)
 
-        tempargs = dict(
-            prefix=b'hg-git-fetch-',
-            suffix=b'.pack',
-            dir=self.gitdir,
-        )
-
-        if self.is_clone and self.gitdir.startswith(self.repo.root):
-            # if we're in a clone, and the git directory is within the
-            # repository just created, so we can use a named temporary
-            # file, suitable for moving into the git repository
-            move = delete = True
-            f = tempfile.NamedTemporaryFile(**tempargs, delete=False)
-        else:
-            move = delete = False
-            max_size = self.ui.configint(b'hggit', b'fetchbuffer') * 1e6
-            f = tempfile.SpooledTemporaryFile(**tempargs, max_size=max_size)
-
         try:
-            ret = self._call_client(
-                remote,
-                'fetch_pack',
-                determine_wants,
-                graphwalker,
-                f.write,
-                progress.progress,
-            )
-
-            if f.tell() != 0:
-                if move:
-                    self.ui.debug(b'moving git pack into %s\n' % self.gitdir)
-                    # windows might have issues moving an open file?
-                    f.close()
-                    self.git.object_store.move_in_pack(f.name)
-                    delete = False
-                else:
-                    self.ui.debug(b'adding git pack to %s\n' % self.gitdir)
-                    f.seek(0)
-                    self.git.object_store.add_thin_pack(f.read, None)
+            with util.add_pack(self.git.object_store) as f:
+                ret = self._call_client(
+                    remote,
+                    'fetch_pack',
+                    determine_wants,
+                    graphwalker,
+                    f.write,
+                    progress.progress,
+                )
 
             # For empty repos dulwich gives us None, but since later
             # we want to iterate over this, we really want an empty
@@ -1555,11 +1653,11 @@ class GitHandler(object):
             )
         finally:
             progress.flush()
-            f.close()
-            if delete:
-                os.remove(f.name)
 
     def _call_client(self, remote, method, *args, **kwargs):
+        if not isinstance(remote, bytes):
+            remote = remote.loc
+
         if remote in self._clients:
             clientobj, path = self._clients[remote]
             return getattr(clientobj, method)(path, *args, **kwargs)
@@ -1576,10 +1674,10 @@ class GitHandler(object):
 
                 return ret
 
-            except (compat.HTTPUnauthorized, GitProtocolError) as e:
+            except (HTTPUnauthorized, GitProtocolError) as e:
                 self.ui.traceback()
 
-                if isinstance(e, compat.HTTPUnauthorized):
+                if isinstance(e, HTTPUnauthorized):
                     # this is a fallback just in case the header isn't
                     # specified
                     self._http_auth_realm = 'Git'
@@ -1592,9 +1690,6 @@ class GitHandler(object):
                     raise error.Abort(
                         b'HTTP proxy requires authentication',
                     )
-                # dulwich 0.19
-                elif 'unexpected http resp 401' in e.args[0]:
-                    self._http_auth_realm = 'Git'
                 else:
                     raise
 
@@ -1636,11 +1731,13 @@ class GitHandler(object):
                 # prior to 0.20.22, dulwich couldn't handle refs
                 # pointing to missing objects, so don't add them
                 if git_sha and git_sha in self.git:
-                    self.git.refs[git_ref] = git_sha
+                    util.set_refs(self.ui, self.git, {git_ref: git_sha})
 
         return exportable
 
     def export_hg_tags(self):
+        new_refs = {}
+
         for tag, sha in self.repo.tags().items():
             if self.repo.tagtype(tag) in (b'global', b'git'):
                 tag = tag.replace(b' ', b'_')
@@ -1665,12 +1762,15 @@ class GitHandler(object):
 
                 # check whether the tag already exists and is
                 # annotated
-                if tag_refname in self.git.refs:
+                if util.ref_exists(tag_refname, self.git.refs):
                     reftarget = self.git.refs[tag_refname]
                     try:
                         peeledtarget = self.git.get_peeled(tag_refname)
-                    except KeyError as e:
-                        peeledtarget = e.args[0]
+                    except KeyError:
+                        self.ui.note(
+                            b'note: failed to peel tag %s' % (tag_refname)
+                        )
+                        peeledtarget = None
 
                     if peeledtarget != reftarget:
                         # warn the user if they tried changing the tag
@@ -1684,9 +1784,12 @@ class GitHandler(object):
                         # otherwise it'd happen on every pull
                         target = reftarget
 
-                self.git.refs[tag_refname] = target
+                new_refs[tag_refname] = target
                 self.tags[tag] = hex(sha)
 
+        if new_refs:
+            util.set_refs(self.ui, self.git, new_refs)
+
     def get_filtered_bookmarks(self):
         bms = self.repo._bookmarks
 
@@ -1703,17 +1806,6 @@ class GitHandler(object):
             return [(_filter_bm(bm), bm, n) for bm, n in bms.items()]
 
     def get_exportable(self):
-        class heads_tags(object):
-            def __init__(self):
-                self.heads = set()
-                self.tags = set()
-
-            def __iter__(self):
-                return itertools.chain(self.heads, self.tags)
-
-            def __bool__(self):
-                return bool(self.heads) or bool(self.tags)
-
         res = collections.defaultdict(heads_tags)
 
         for filtered_bm, bm, node in self.get_filtered_bookmarks():
@@ -1741,19 +1833,18 @@ class GitHandler(object):
             return
         repotags = self.repo.tags()
         for k in refs:
-            ref_name = k
-            parts = k.split(b'/')
-            if parts[0] == b'refs' and parts[1] == b'tags':
-                ref_name = b"/".join(v for v in parts[2:])
+            if k.startswith(LOCAL_TAG_PREFIX):
+                ref_name = k[len(LOCAL_TAG_PREFIX) :]
+
                 # refs contains all the refs in the server, not just
                 # the ones we are pulling
                 if refs[k] not in self.git.object_store:
                     continue
-                if ref_name[-3:] == ANNOTATED_TAG_SUFFIX:
-                    ref_name = ref_name[:-3]
+                if ref_name.endswith(ANNOTATED_TAG_SUFFIX):
+                    continue
                 if ref_name not in repotags:
                     sha = self.map_hg_get(refs[k], deref=True)
-                    if sha is not None:
+                    if sha is not None and sha is not None:
                         self.tags[ref_name] = sha
 
         self.save_tags()
@@ -1855,12 +1946,12 @@ class GitHandler(object):
 
                 # only log additions on subsequent pulls
                 if not self.is_clone:
-                    self.ui.status(_("adding bookmark %s\n") % bm)
+                    self.ui.status(_(b"adding bookmark %s\n") % bm)
 
             elif unfiltered[current_node].isancestorof(unfiltered[wanted_node]):
                 # fast forward
                 changes.append((bm, wanted_node))
-                self.ui.status(_("updating bookmark %s\n") % bm)
+                self.ui.status(_(b"updating bookmark %s\n") % bm)
 
             elif unfiltered.obsstore and wanted_node in obsutil.foreground(
                 unfiltered, [current_node]
@@ -1870,11 +1961,11 @@ class GitHandler(object):
                 # a background thingy that is more efficient that
                 # the foreground one.)
                 changes.append((bm, wanted_node))
-                self.ui.status(_("updating bookmark %s\n") % bm)
+                self.ui.status(_(b"updating bookmark %s\n") % bm)
 
             else:
                 self.ui.status(
-                    _("not updating diverged bookmark %s\n") % bm,
+                    _(b"not updating diverged bookmark %s\n") % bm,
                 )
 
         if changes:
@@ -1882,17 +1973,7 @@ class GitHandler(object):
                 with self.repo.transaction(b"hg-git") as tr:
                     bms.applychanges(self.repo, tr, changes)
 
-    def get_public_heads(self, remote_name, refs):
-        nodeids_to_publish = set()
-
-        for sha in git2hg.get_public(self.ui, refs, remote_name):
-            hgsha = self.map_hg_get(sha, deref=True)
-            if hgsha:
-                nodeids_to_publish.add(bin(hgsha))
-
-        return nodeids_to_publish
-
-    def update_remote_branches(self, remote_name, refs):
+    def _update_remote_branches_for(self, remote_name, refs):
         remote_refs = self.remote_refs
 
         if self.ui.configbool(b'git', b'pull-prune-remote-branches'):
@@ -1912,30 +1993,36 @@ class GitHandler(object):
         all_remote_nodeids = []
 
         for ref_name, sha in refs.items():
-            hgsha = self.map_hg_get(sha)
+            if ref_name.endswith(ANNOTATED_TAG_SUFFIX):
+                # the sha points to a peeled tag; we should either
+                # pick it up through the tag itself, or ignore it
+                continue
+
+            hgsha = self.map_hg_get(sha, deref=True)
 
             if (
                 ref_name.startswith(LOCAL_BRANCH_PREFIX)
                 and hgsha is not None
                 and hgsha in self.repo
             ):
-                head = ref_name[11:]
+                head = ref_name[len(LOCAL_BRANCH_PREFIX) :]
                 remote_head = b'/'.join((remote_name, head))
 
                 remote_refs[remote_head] = bin(hgsha)
                 # TODO(durin42): what is this doing?
                 new_ref = REMOTE_BRANCH_PREFIX + remote_head
-                self.git.refs[new_ref] = sha
-            elif ref_name.startswith(
-                LOCAL_TAG_PREFIX
-            ) and not ref_name.endswith(ANNOTATED_TAG_SUFFIX):
-                self.git.refs[ref_name] = sha
+
+                util.set_refs(self.ui, self.git, {new_ref: sha})
+            elif ref_name.startswith(LOCAL_TAG_PREFIX):
+                util.set_refs(self.ui, self.git, {ref_name: sha})
 
             if hgsha:
                 all_remote_nodeids.append(bin(hgsha))
 
-        with self.repo.lock(), self.repo.transaction(b"hg-git-phases") as tr:
-            if all_remote_nodeids:
+        if all_remote_nodeids:
+            with self.repo.lock(), self.repo.transaction(
+                b"hg-git-phases"
+            ) as tr:
                 # sanity check: ensure that all corresponding commits
                 # are at least draft; this can happen on no-op pulls
                 # where the commit already exists, but is secret
@@ -1945,12 +2032,25 @@ class GitHandler(object):
                     phases.draft,
                     all_remote_nodeids,
                 )
-            # ensure that we update phases on push and no-op pulls
+
+    def update_remote_branches(self, remote_names, refs):
+        for remote_name in remote_names:
+            self._update_remote_branches_for(remote_name, refs)
+
+        # ensure that we update phases on push and no-op pulls
+        with self.repo.lock(), self.repo.transaction(b"hg-git-phases") as tr:
+            nodeids_to_publish = set()
+
+            for sha in git2hg.get_public(self.ui, refs, remote_names):
+                hgsha = self.map_hg_get(sha, deref=True)
+                if hgsha:
+                    nodeids_to_publish.add(bin(hgsha))
+
             phases.advanceboundary(
                 self.repo,
                 tr,
                 phases.public,
-                self.get_public_heads(remote_name, refs),
+                nodeids_to_publish,
             )
 
     # UTILITY FUNCTIONS
@@ -2092,22 +2192,24 @@ class GitHandler(object):
         )
 
     def remote_names(self, remote, push):
+        if not isinstance(remote, bytes):
+            return [remote.name] if remote.name is not None else []
+
         names = set()
-        url = compat.url(remote)
+        url = urlutil.url(remote)
 
         if url.islocal() and not url.isabs():
             remote = os.path.abspath(url.localpath())
 
         for name, paths in self.ui.paths.items():
-            # paths became lists in mercurial 5.9
-            if not isinstance(paths, list):
-                paths = [paths]
-
             for path in paths:
                 # ignore aliases
                 if hasattr(path, 'raw_url') and path.raw_url.scheme == b'path':
                     continue
-                loc = push and path.pushloc or path.loc
+                if push:
+                    loc = compat.get_push_location(path)
+                else:
+                    loc = path.loc
                 if loc == remote:
                     names.add(name)
 
@@ -2124,7 +2226,7 @@ class GitHandler(object):
             if opt == b'abort':
                 raise error.Abort(
                     b"invalid path '%s' rejected by configuration" % prettypath,
-                    hint=b"see 'hg help hggit' for details",
+                    hint=b"see 'hg help config.hggit.invalidpaths for details",
                 )
             elif opt == b'keep' and b'\r' not in path and b'\n' not in path:
                 ui.warn(
@@ -2217,16 +2319,12 @@ class GitHandler(object):
                     )
 
             str_uri = uri.decode('utf-8')
-            # not available in dulwich 0.19
-            if hasattr(client, 'get_credentials_from_store'):
-                urlobj = compat.url(uri)
-                auth = client.get_credentials_from_store(
-                    urlobj.scheme,
-                    urlobj.host,
-                    urlobj.user,
-                )
-            else:
-                auth = None
+            urlobj = urlutil.url(uri)
+            auth = client.get_credentials_from_store(
+                urlobj.scheme,
+                urlobj.host,
+                urlobj.user,
+            )
 
             if self._http_auth_realm:
                 # since we've tried an unauthenticated request, and
@@ -2261,7 +2359,7 @@ class GitHandler(object):
             )
 
         if uri.startswith(b'file://'):
-            return client.LocalGitClient(), compat.url(uri).path
+            return client.LocalGitClient(), urlutil.url(uri).path
 
         # if its not git or git+ssh, try a local url..
         return client.SubprocessGitClient(), uri
